from typing import Union

from multipledispatch import dispatch

from ...expr.prim_ops import Add, Sub, Mul, Div, Rem, \
    Lt, Leq, Gt, Geq, Eq, Neq, Xor, Or, And, Not, Neg, \
    Cat, Bits, AsUInt, AsSInt, Shl, Shr, Dshl, Dshr
from ...type import UIntType, SIntType, ClockType
from ...type_checker.utils import type_in, check_all_same_uint_sint

checker = dispatch


###############################################################
#                      TYPE CHECKERS                          #
###############################################################

@checker(Add)
def check(add: Add):
    from .. import check_all_expr
    if not check_all_expr(*add.args):
        return False

    if not check_all_same_uint_sint(add.args[0].tpe,
                                    add.args[1].tpe,
                                    add.tpe):
        return False

    expected_type_width = max(add.args[0].tpe.width.width,
                              add.args[1].tpe.width.width) + 1
    if add.tpe.width.width != expected_type_width:
        return False

    return True


@checker(Sub)
def check(sub: Sub):
    from .. import check_all_expr
    if not check_all_expr(*sub.args):
        return False

    if not check_all_same_uint_sint(sub.args[0].tpe,
                                    sub.args[1].tpe):
        return False

    if not type_in(sub.tpe, SIntType):
        return False

    expected_type_width = max(sub.args[0].tpe.width.width,
                              sub.args[1].tpe.width.width) + 1
    if sub.tpe.width.width != expected_type_width:
        return False

    return True


@checker(Mul)
def check(mul: Mul):
    from .. import check_all_expr
    if not check_all_expr(*mul.args):
        return False

    if not check_all_same_uint_sint(mul.args[0].tpe,
                                    mul.args[1].tpe,
                                    mul.tpe):
        return False

    expected_type_width = \
        mul.args[0].tpe.width.width + mul.args[1].tpe.width.width
    if mul.tpe.width.width != expected_type_width:
        return False

    return True


@checker(Div)
def check(div: Div):
    from .. import check_all_expr
    if not check_all_expr(*div.args):
        return False

    if not check_all_same_uint_sint(div.args[0].tpe,
                                    div.args[1].tpe,
                                    div.tpe):
        return False

    expected_type_width = div.args[0].tpe.width.width
    if type_in(div.args[0].tpe, SIntType):
        expected_type_width += 1
    if div.tpe.width.width != expected_type_width:
        return False

    return True


@checker(Rem)
def check(rem: Rem):
    from .. import check_all_expr
    if not check_all_expr(*rem.args):
        return False

    if not check_all_same_uint_sint(rem.args[0].tpe,
                                    rem.args[1].tpe,
                                    rem.tpe):
        return False

    expected_type_width = min(rem.args[0].tpe.width.width,
                              rem.args[1].tpe.width.width)
    if rem.tpe.width.width != expected_type_width:
        return False

    return True


@checker((Lt, Leq, Gt, Geq, Eq, Neq))
def check(comparison: Union[Lt, Leq, Gt, Geq, Eq, Neq]):
    from .. import check_all_expr
    if not check_all_expr(*comparison.args):
        return False

    if not check_all_same_uint_sint(comparison.args[0].tpe,
                                    comparison.args[1].tpe):
        return False

    if not type_in(comparison.tpe, UIntType):
        return False

    if comparison.tpe.width.width != 1:
        return False

    return True


@checker((And, Or, Xor))
def check(binary_bit: Union[And, Or, Xor]):
    from .. import check_all_expr
    if not check_all_expr(*binary_bit.args):
        return False

    if not check_all_same_uint_sint(binary_bit.args[0].tpe,
                                    binary_bit.args[1].tpe):
        return False

    if not type_in(binary_bit.tpe, UIntType):
        return False

    expected_type_width = max(binary_bit.args[0].tpe.width.width,
                              binary_bit.args[1].tpe.width.width)
    if binary_bit.tpe.width.width != expected_type_width:
        return False

    return True


@checker(Not)
def check(n: Not):
    from .. import check_all_expr
    if not check_all_expr(n.arg):
        return False

    if not check_all_same_uint_sint(n.arg.tpe):
        return False

    if not type_in(n.tpe, UIntType):
        return False

    expected_type_width = n.arg.tpe.width.width
    if n.tpe.width.width != expected_type_width:
        return False

    return True


@checker(Neg)
def check(neg: Neg):
    from .. import check_all_expr
    if not check_all_expr(neg.arg):
        return False

    if not check_all_same_uint_sint(neg.arg.tpe):
        return False

    if not type_in(neg.tpe, SIntType):
        return False

    expected_type_width = neg.arg.tpe.width.width + 1
    if neg.tpe.width.width != expected_type_width:
        return False

    return True


@checker(Cat)
def check(cat: Cat):
    from .. import check_all_expr
    if not check_all_expr(*cat.args):
        return False

    if not check_all_same_uint_sint(cat.args[0].tpe,
                                    cat.args[1].tpe):
        return False

    if not type_in(cat.tpe, UIntType):
        return False

    expected_type_width = \
        cat.args[0].tpe.width.width + cat.args[1].tpe.width.width
    if cat.tpe.width.width != expected_type_width:
        return False

    return True


@checker(Bits)
def check(bits: Bits):
    from .. import check_all_expr
    if not check_all_expr(bits.ir_arg):
        return False

    if not check_all_same_uint_sint(bits.ir_arg.tpe):
        return False

    if not type_in(bits.tpe, UIntType):
        return False

    if not \
            bits.ir_arg.tpe.width.width > \
            bits.const_args[0] >= \
            bits.const_args[1] >= 0:
        return False

    expected_type_width = bits.const_args[0] - bits.const_args[1] + 1
    if bits.tpe.width.width != expected_type_width:
        return False

    return True


@checker(AsUInt)
def check(as_uint: AsUInt):
    from .. import check_all_expr
    if not check_all_expr(as_uint.arg):
        return False

    if not type_in(as_uint.arg.tpe, UIntType, SIntType, ClockType):
        return False

    if not type_in(as_uint.tpe, UIntType):
        return False

    if type_in(as_uint.arg.tpe, ClockType):
        expected_type_width = 1
    else:
        expected_type_width = as_uint.arg.tpe.width.width

    if as_uint.tpe.width.width != expected_type_width:
        return False

    return True


@checker(AsSInt)
def check(as_sint: AsSInt):
    from .. import check_all_expr
    if not check_all_expr(as_sint.arg):
        return False

    if not type_in(as_sint.arg.tpe, UIntType, SIntType, ClockType):
        return False

    if not type_in(as_sint.tpe, SIntType):
        return False

    if type_in(as_sint.arg.tpe, ClockType):
        expected_type_width = 1
    else:
        expected_type_width = as_sint.arg.tpe.width.width

    if as_sint.tpe.width.width != expected_type_width:
        return False

    return True


@checker(Shl)
def check(shl: Shl):
    from .. import check_all_expr
    if not check_all_expr(shl.ir_arg):
        return False

    if not check_all_same_uint_sint(shl.ir_arg.tpe, shl.tpe):
        return False

    expected_type_width = shl.ir_arg.tpe.width.width + shl.const_arg
    if shl.tpe.width.width != expected_type_width:
        return False

    return True


@checker(Shr)
def check(shr: Shr):
    from .. import check_all_expr
    if not check_all_expr(shr.ir_arg):
        return False

    if not check_all_same_uint_sint(shr.ir_arg.tpe, shr.tpe):
        return False

    expected_type_width = shr.ir_arg.tpe.width.width - shr.const_arg
    expected_type_width = max(expected_type_width, 1)
    if shr.tpe.width.width != expected_type_width:
        return False

    return True


@checker(Dshl)
def check(dshl: Dshl):
    from .. import check_all_expr
    if not check_all_expr(*dshl.args):
        return False

    if not check_all_same_uint_sint(dshl.args[0].tpe,
                                    dshl.tpe):
        return False

    if not type_in(dshl.args[1].tpe, UIntType):
        return False

    expected_type_width = \
        dshl.args[0].tpe.width.width + 2 ** dshl.args[1].tpe.width.width - 1
    if dshl.tpe.width.width != expected_type_width:
        return False

    return True


@checker(Dshr)
def check(dshr: Dshr):
    from .. import check_all_expr
    if not check_all_expr(*dshr.args):
        return False

    if not check_all_same_uint_sint(dshr.args[0].tpe,
                                    dshr.tpe):
        return False

    if not type_in(dshr.args[1].tpe, UIntType):
        return False

    expected_type_width = dshr.args[0].tpe.width.width
    if dshr.tpe.width.width != expected_type_width:
        return False

    return True



from ..utils import serialize_str
from . import Statement


class Conditionally(Statement):
    def __init__(self, pred_ref, seq, alt):
        self.pred_ref = pred_ref
        self.seq = seq
        self.alt = alt

    def serialize_stmt(self, output, indent):
        output.write(b"when ")
        self.pred_ref.serialize(output)
        output.write(b" :\n")
        indent += 1
        output.write(serialize_str("  " * indent))
        self.seq.serialize_stmt(output, indent)
        output.write(b"\n")
        output.write(serialize_str("  " * (indent - 1)))
        output.write(b"else :\n")
        output.write(serialize_str("  " * indent))
        self.alt.serialize_stmt(output, indent)



from . import Statement


class Connect(Statement):
    def __init__(self, loc_ref, expr_ref):
        self.loc_ref = loc_ref
        self.expr_ref = expr_ref

    def serialize_stmt(self, output, indent):
        self.loc_ref.serialize(output)
        output.write(b" <= ")
        self.expr_ref.serialize(output)



from ...utils import serialize_str
from .. import Statement


class DefMemory(Statement):
    def __init__(self, name, tpe):
        self.name = name
        self.tpe = tpe

    def serialize_stmt(self, output, indent):
        output.write(b"cmem ")
        output.write(serialize_str(self.name))
        output.write(b" : ")
        self.tpe.serialize(output)


class DefMemReadPort(Statement):
    def __init__(self, name, mem_ref, index_ref, clock_ref):
        self.name = name
        self.mem_ref = mem_ref
        self.index_ref = index_ref
        self.clock_ref = clock_ref

    def serialize_stmt(self, output, indent):
        output.write(b"read mport ")
        output.write(serialize_str(self.name))
        output.write(b" = ")
        self.mem_ref.serialize(output)
        output.write(b"[")
        self.index_ref.serialize(output)
        output.write(b"], ")
        self.clock_ref.serialize(output)


class DefMemWritePort(Statement):
    def __init__(self, name, mem_ref, index_ref, clock_ref):
        self.name = name
        self.mem_ref = mem_ref
        self.index_ref = index_ref
        self.clock_ref = clock_ref

    def serialize_stmt(self, output, indent):
        output.write(b"write mport ")
        output.write(serialize_str(self.name))
        output.write(b" = ")
        self.mem_ref.serialize(output)
        output.write(b"[")
        self.index_ref.serialize(output)
        output.write(b"], ")
        self.clock_ref.serialize(output)



from py_hcl.firrtl_ir.expr.prim_ops import Bits
from py_hcl.firrtl_ir.shortcuts import uw, u, w, s
from py_hcl.firrtl_ir.type import UIntType, SIntType, \
    UnknownType, VectorType, BundleType
from tests.test_firrtl_ir.utils import serialize_equal
from .helper import OpCase, basis_tester, \
    encounter_error_tester, width


def args(*arg_types):
    class A:
        @staticmethod
        def const(*const_types):
            class B:
                @staticmethod
                def filter(valid_fn):
                    class C:
                        @staticmethod
                        def tpe(res_type):
                            return OpCase(Bits) \
                                .arg_types(*arg_types) \
                                .const_arg_types(*const_types) \
                                .filter(valid_fn) \
                                .res_type(res_type)

                    return C

            return B

    return A


bits_basis_cases = [
    args(UIntType).const(int, int).filter(
        lambda u, a, b: width(u) > a >= b >= 0).tpe(
        lambda u, a, b: uw(a - b + 1)),
    args(SIntType).const(int, int).filter(
        lambda u, a, b: width(u) > a >= b >= 0).tpe(
        lambda u, a, b: uw(a - b + 1)),
]

bits_type_wrong_cases = [
    args(UnknownType).const(int, int).filter(
        lambda u, a, b: a >= b >= 0).tpe(
        lambda u, a, b: uw(a - b + 1)),
    args(VectorType).const(int, int).filter(
        lambda u, a, b: a >= b >= 0).tpe(
        lambda u, a, b: uw(a - b + 1)),
    args(BundleType).const(int, int).filter(
        lambda u, a, b: a >= b >= 0).tpe(
        lambda u, a, b: uw(a - b + 1)),
]

bits_width_wrong_cases = [
    args(UIntType).const(int, int).filter(
        lambda u, a, b: width(u) > a >= b >= 0).tpe(
        lambda u, a, b: uw(a - b + 2)),
    args(SIntType).const(int, int).filter(
        lambda u, a, b: width(u) > a >= b >= 0).tpe(
        lambda u, a, b: uw(a - b + 2)),
    args(UIntType).const(int, int).filter(
        lambda u, a, b: width(u) > a >= b >= 0).tpe(
        lambda u, a, b: uw(a - b)),
    args(SIntType).const(int, int).filter(
        lambda u, a, b: width(u) > a >= b >= 0).tpe(
        lambda u, a, b: uw(a - b)),
]

bits_invalid_cases = [
    args(UIntType).const(int, int).filter(
        lambda u, a, b: width(u) <= a and a >= b >= 0).tpe(
        lambda u, a, b: uw(a - b + 1)),
    args(SIntType).const(int, int).filter(
        lambda u, a, b: width(u) <= a and a >= b >= 0).tpe(
        lambda u, a, b: uw(a - b + 1)),
    args(UIntType).const(int, int).filter(
        lambda u, a, b: width(u) > a and b > a >= 0).tpe(
        lambda u, a, b: uw(b - a + 1)),
    args(SIntType).const(int, int).filter(
        lambda u, a, b: width(u) > a and b > a >= 0).tpe(
        lambda u, a, b: uw(b - a + 1)),
]


def test_bits():
    basis_tester(bits_basis_cases)
    encounter_error_tester(bits_type_wrong_cases)
    encounter_error_tester(bits_width_wrong_cases)
    encounter_error_tester(bits_invalid_cases)
    serialize_equal(Bits(u(20, w(5)), [4, 4], uw(1)),
                    'bits(UInt<5>("h14"), 4, 4)')
    serialize_equal(Bits(s(-20, w(6)), [4, 3], uw(2)),
                    'bits(SInt<6>("h-14"), 4, 3)')



from typing import Optional, Dict

from py_hcl.utils import json_serialize


@json_serialize
class NamedExprHolder(object):
    def __init__(self, module_name: str,
                 named_expression_table: Dict[int, str]):
        self.module_name = module_name
        self.named_expression_table = named_expression_table


@json_serialize
class NamedExprNode(object):
    def __init__(self,
                 named_expr_holder: NamedExprHolder,
                 next_node: Optional["NamedExprNode"]):
        self.named_expr_holder = named_expr_holder
        if next_node:
            self.next_node = next_node


@json_serialize
class NamedExprChain(object):
    def __init__(self, named_expr_chain_head: NamedExprNode):
        self.named_expr_chain_head = named_expr_chain_head



from multipledispatch import dispatch

from py_hcl.core.type import HclType


@dispatch()
def invert_exp(self: HclType):
    from py_hcl.core.type.bundle import Dir
    return {'dir': Dir.SINK, 'hcl_type': self}


def bd_fld_wrap(cls):
    if hasattr(cls, '__invert__'):
        return cls

    cls.__invert__ = invert_exp
    return cls


@dispatch()
def vec_exp(self: HclType, i: int):
    from py_hcl.core.type.vector import VectorT
    return VectorT(self, i)


@dispatch()
def vec_exp(self: HclType, t: tuple):
    from py_hcl.core.type.vector import VectorT

    # TODO: Accurate Error Message
    assert all(isinstance(i, int) for i in t)
    v = self
    for s in t[::-1]:
        v = VectorT(v, s)
    return v


def vec_wrap(cls):
    if hasattr(cls, '__getitem__'):
        return cls

    cls.__getitem__ = vec_exp
    return cls



from .branch import when, else_when, otherwise  # noqa: F401
from .module import Module  # noqa: F401
from .tpe.clock import Clock  # noqa: F401
from .tpe.uint import U, Bool  # noqa: F401
from .tpe.sint import S  # noqa: F401
from .tpe.bundle import Bundle  # noqa: F401
from .expr.io import IO, Input, Output, io_extend  # noqa: F401
from .expr.wire import Wire  # noqa: F401



from typing import Tuple, List


def module_inherit_mro(modules: Tuple[type]) -> List[type]:
    from py_hcl.core.module.meta_module import MetaModule

    modules = type("_hcl_fake_module", modules, {}).mro()
    modules = [m for m in modules[1:] if isinstance(m, MetaModule)]
    return modules



from py_hcl.firrtl_ir.expr.mux import Mux
from py_hcl.firrtl_ir.shortcuts import n, uw, u, vec, sw, w
from py_hcl.firrtl_ir.type_checker import check
from ..utils import serialize_equal


def test_mux_basis():
    mux = Mux(n("c", uw(1)), n("a", uw(8)), n("b", uw(8)), uw(8))
    assert check(mux)
    serialize_equal(mux, "mux(c, a, b)")

    mux = Mux(u(1, w(1)),
              n("b", vec(sw(8), 10)),
              n("c", vec(sw(8), 10)),
              vec(sw(8), 10))
    assert check(mux)
    serialize_equal(mux, 'mux(UInt<1>("h1"), b, c)')


def test_mux_cond_type_wrong():
    mux = Mux(n("c", uw(2)), n("a", uw(8)), n("b", uw(8)), uw(8))
    assert not check(mux)

    mux = Mux(n("c", sw(1)), n("a", uw(8)), n("b", uw(8)), uw(8))
    assert not check(mux)

    mux = Mux(n("c", vec(uw(1), 1)), n("a", uw(8)), n("b", uw(8)), uw(8))
    assert not check(mux)


def test_mux_tf_value_type_wrong():
    mux = Mux(n("c", uw(1)), n("a", uw(7)), n("b", uw(8)), uw(8))
    assert not check(mux)

    mux = Mux(n("c", uw(1)), n("a", uw(8)), n("b", sw(8)), uw(8))
    assert not check(mux)

    mux = Mux(n("c", uw(1)), n("a", uw(8)), n("b", uw(8)), sw(8))
    assert not check(mux)



from . import Expression
from ..utils import serialize_num


class Add(Expression):
    def __init__(self, args, tpe):
        self.args = args
        self.tpe = tpe

    def serialize(self, output):
        output.write(b"add(")
        self.args[0].serialize(output)
        output.write(b", ")
        self.args[1].serialize(output)
        output.write(b")")


class Sub(Expression):
    def __init__(self, args, tpe):
        self.args = args
        self.tpe = tpe

    def serialize(self, output):
        output.write(b"sub(")
        self.args[0].serialize(output)
        output.write(b", ")
        self.args[1].serialize(output)
        output.write(b")")


class Mul(Expression):
    def __init__(self, args, tpe):
        self.args = args
        self.tpe = tpe

    def serialize(self, output):
        output.write(b"mul(")
        self.args[0].serialize(output)
        output.write(b", ")
        self.args[1].serialize(output)
        output.write(b")")


class Div(Expression):
    def __init__(self, args, tpe):
        self.args = args
        self.tpe = tpe

    def serialize(self, output):
        output.write(b"div(")
        self.args[0].serialize(output)
        output.write(b", ")
        self.args[1].serialize(output)
        output.write(b")")


class Rem(Expression):
    def __init__(self, args, tpe):
        self.args = args
        self.tpe = tpe

    def serialize(self, output):
        output.write(b"rem(")
        self.args[0].serialize(output)
        output.write(b", ")
        self.args[1].serialize(output)
        output.write(b")")


class Lt(Expression):
    def __init__(self, args, tpe):
        self.args = args
        self.tpe = tpe

    def serialize(self, output):
        output.write(b"lt(")
        self.args[0].serialize(output)
        output.write(b", ")
        self.args[1].serialize(output)
        output.write(b")")


class Leq(Expression):
    def __init__(self, args, tpe):
        self.args = args
        self.tpe = tpe

    def serialize(self, output):
        output.write(b"leq(")
        self.args[0].serialize(output)
        output.write(b", ")
        self.args[1].serialize(output)
        output.write(b")")


class Gt(Expression):
    def __init__(self, args, tpe):
        self.args = args
        self.tpe = tpe

    def serialize(self, output):
        output.write(b"gt(")
        self.args[0].serialize(output)
        output.write(b", ")
        self.args[1].serialize(output)
        output.write(b")")


class Geq(Expression):
    def __init__(self, args, tpe):
        self.args = args
        self.tpe = tpe

    def serialize(self, output):
        output.write(b"geq(")
        self.args[0].serialize(output)
        output.write(b", ")
        self.args[1].serialize(output)
        output.write(b")")


class Eq(Expression):
    def __init__(self, args, tpe):
        self.args = args
        self.tpe = tpe

    def serialize(self, output):
        output.write(b"eq(")
        self.args[0].serialize(output)
        output.write(b", ")
        self.args[1].serialize(output)
        output.write(b")")


class Neq(Expression):
    def __init__(self, args, tpe):
        self.args = args
        self.tpe = tpe

    def serialize(self, output):
        output.write(b"neq(")
        self.args[0].serialize(output)
        output.write(b", ")
        self.args[1].serialize(output)
        output.write(b")")


class And(Expression):
    def __init__(self, args, tpe):
        self.args = args
        self.tpe = tpe

    def serialize(self, output):
        output.write(b"and(")
        self.args[0].serialize(output)
        output.write(b", ")
        self.args[1].serialize(output)
        output.write(b")")


class Or(Expression):
    def __init__(self, args, tpe):
        self.args = args
        self.tpe = tpe

    def serialize(self, output):
        output.write(b"or(")
        self.args[0].serialize(output)
        output.write(b", ")
        self.args[1].serialize(output)
        output.write(b")")


class Xor(Expression):
    def __init__(self, args, tpe):
        self.args = args
        self.tpe = tpe

    def serialize(self, output):
        output.write(b"xor(")
        self.args[0].serialize(output)
        output.write(b", ")
        self.args[1].serialize(output)
        output.write(b")")


class Not(Expression):
    def __init__(self, arg, tpe):
        self.arg = arg
        self.tpe = tpe

    def serialize(self, output):
        output.write(b"not(")
        self.arg.serialize(output)
        output.write(b")")


class Neg(Expression):
    def __init__(self, arg, tpe):
        self.arg = arg
        self.tpe = tpe

    def serialize(self, output):
        output.write(b"neg(")
        self.arg.serialize(output)
        output.write(b")")


class Cat(Expression):
    def __init__(self, args, tpe):
        self.args = args
        self.tpe = tpe

    def serialize(self, output):
        output.write(b"cat(")
        self.args[0].serialize(output)
        output.write(b", ")
        self.args[1].serialize(output)
        output.write(b")")


class Bits(Expression):
    def __init__(self, ir_arg, const_args, tpe):
        self.ir_arg = ir_arg
        self.const_args = const_args
        self.tpe = tpe

    def serialize(self, output):
        output.write(b"bits(")
        self.ir_arg.serialize(output)
        output.write(b", ")
        output.write(serialize_num(self.const_args[0]))
        output.write(b", ")
        output.write(serialize_num(self.const_args[1]))
        output.write(b")")


class AsUInt(Expression):
    def __init__(self, arg, tpe):
        self.arg = arg
        self.tpe = tpe

    def serialize(self, output):
        output.write(b"asUInt(")
        self.arg.serialize(output)
        output.write(b")")


class AsSInt(Expression):
    def __init__(self, arg, tpe):
        self.arg = arg
        self.tpe = tpe

    def serialize(self, output):
        output.write(b"asSInt(")
        self.arg.serialize(output)
        output.write(b")")


class Shl(Expression):
    def __init__(self, ir_arg, const_arg, tpe):
        self.ir_arg = ir_arg
        self.const_arg = const_arg
        self.tpe = tpe

    def serialize(self, output):
        output.write(b"shl(")
        self.ir_arg.serialize(output)
        output.write(b", ")
        output.write(serialize_num(self.const_arg))
        output.write(b")")


class Shr(Expression):
    def __init__(self, ir_arg, const_arg, tpe):
        self.ir_arg = ir_arg
        self.const_arg = const_arg
        self.tpe = tpe

    def serialize(self, output):
        output.write(b"shr(")
        self.ir_arg.serialize(output)
        output.write(b", ")
        output.write(serialize_num(self.const_arg))
        output.write(b")")


class Dshl(Expression):
    def __init__(self, args, tpe):
        self.args = args
        self.tpe = tpe

    def serialize(self, output):
        output.write(b"dshl(")
        self.args[0].serialize(output)
        output.write(b", ")
        self.args[1].serialize(output)
        output.write(b")")


class Dshr(Expression):
    def __init__(self, args, tpe):
        self.args = args
        self.tpe = tpe

    def serialize(self, output):
        output.write(b"dshr(")
        self.args[0].serialize(output)
        output.write(b", ")
        self.args[1].serialize(output)
        output.write(b")")



from typing import Dict, Union, List

from py_hcl.transformer.pyhcl_to_firrtl.conv_type import convert_type
from py_hcl.core.expr.io import Input, Output
from py_hcl.firrtl_ir.stmt.defn.module import InputPort, OutputPort
from py_hcl.firrtl_ir.type import ClockType, UIntType, BundleType
from py_hcl.firrtl_ir.type.field import Field
from py_hcl.firrtl_ir.type.width import Width


def convert_ports(raw_ports: Dict[str, Union[Input, Output]]):
    ports = [InputPort('clock', ClockType()),
             InputPort('reset', UIntType(Width(1)))]
    for k, v in raw_ports.items():
        p = InputPort if v.port_dir == 'input' else OutputPort
        ports.append(p(k, convert_type(v.hcl_type)))
    return ports


def ports_to_bundle_type(ports: List[Union[InputPort, OutputPort]]):
    fields = []
    for p in ports:
        if isinstance(p, InputPort):
            fields.append(Field(p.name, p.tpe, True))
        else:
            fields.append(Field(p.name, p.tpe, False))
    return BundleType(fields)



from py_hcl.transformer.pyhcl_to_firrtl.context import Context


def build_io_name(module_name: str, field_name: str):
    return module_name + "_io_" + field_name


def build_reserve_name(module_name: str, expr_name: str):
    return "_" + module_name + "_" + expr_name


def get_io_obj(packed_module):
    table = packed_module.named_expr_chain.named_expr_chain_head \
        .named_expr_holder.named_expression_table
    io_id = list(table.keys())[list(table.values()).index('io')]
    return Context.expr_table[io_id]



from py_hcl.core.stmt.connect import Connect
from py_hcl.dsl.expr.io import IO
from py_hcl.dsl.expr.wire import Wire
from py_hcl.dsl.module import Module
from py_hcl.dsl.tpe.uint import U


def test_statement():
    class A(Module):
        io = IO()

        a = Wire(U.w(8))
        b = Wire(U.w(8))
        c = Wire(U.w(8))

        c <<= a + b

    s = A.packed_module.statement_chain \
        .stmt_chain_head.stmt_holder.top_statement.statements
    assert len(s) == 1
    assert isinstance(s[0].statement, Connect)



from py_hcl.firrtl_ir.expr.prim_ops import Sub
from py_hcl.firrtl_ir.shortcuts import sw, u, w, s
from py_hcl.firrtl_ir.type import SIntType, UIntType
from tests.test_firrtl_ir.utils import serialize_equal
from .helper import basis_tester, encounter_error_tester, OpCase, \
    type_wrong_cases_2_args_gen


def max_width(x, y):
    return max(x.tpe.width.width, y.tpe.width.width)


def args(*arg_types):
    class C:
        @staticmethod
        def tpe(res_type):
            return OpCase(Sub).arg_types(*arg_types).res_type(res_type)

    return C


sub_basis_cases = [
    args(UIntType, UIntType).tpe(lambda x, y: sw(max_width(x, y) + 1)),
    args(SIntType, SIntType).tpe(lambda x, y: sw(max_width(x, y) + 1)),
]

sub_type_wrong_cases = type_wrong_cases_2_args_gen(Sub)

sub_width_wrong_cases = [
    args(UIntType, UIntType).tpe(lambda x, y: sw(max_width(x, y))),
    args(SIntType, SIntType).tpe(lambda x, y: sw(max_width(x, y))),
    args(UIntType, UIntType).tpe(lambda x, y: sw(max_width(x, y) + 2)),
    args(SIntType, SIntType).tpe(lambda x, y: sw(max_width(x, y) + 2)),
    args(UIntType, UIntType).tpe(lambda x, y: sw(max_width(x, y) - 1)),
    args(SIntType, SIntType).tpe(lambda x, y: sw(max_width(x, y) - 1)),
    args(UIntType, UIntType).tpe(lambda x, y: sw(1)),
    args(SIntType, SIntType).tpe(lambda x, y: sw(1)),
]


def test_sub():
    basis_tester(sub_basis_cases)
    encounter_error_tester(sub_type_wrong_cases)
    encounter_error_tester(sub_width_wrong_cases)
    serialize_equal(Sub([u(20, w(5)), u(15, w(4))], sw(6)),
                    'sub(UInt<5>("h14"), UInt<4>("hf"))')
    serialize_equal(Sub([s(-20, w(6)), s(-15, w(5))], sw(7)),
                    'sub(SInt<6>("h-14"), SInt<5>("h-f"))')



from py_hcl.firrtl_ir.shortcuts import uw, n, w, u, s
from py_hcl.firrtl_ir.stmt.block import Block
from py_hcl.firrtl_ir.stmt.conditionally import Conditionally
from py_hcl.firrtl_ir.stmt.connect import Connect
from py_hcl.firrtl_ir.stmt.defn.module import DefModule, \
    OutputPort, InputPort, DefExtModule
from py_hcl.firrtl_ir.stmt.defn.node import DefNode
from py_hcl.firrtl_ir.stmt.empty import EmptyStmt
from py_hcl.firrtl_ir.type_checker import check
from ...utils import serialize_stmt_equal


def test_module_basis():
    mod = DefModule("m", [OutputPort("p", uw(8))],
                    Connect(n("p", uw(8)), u(2, w(8))))
    assert check(mod)
    serialize_stmt_equal(mod, 'module m :\n'
                              '  output p : UInt<8>\n'
                              '\n'
                              '  p <= UInt<8>("h2")')

    mod = DefModule("m", [InputPort("b", uw(8)),
                          OutputPort("a", uw(8))],
                    Block([DefNode("n", u(1, w(1))),
                           Conditionally(n("n", uw(1)),
                                         EmptyStmt(),
                                         Connect(n("a", uw(8)), n("b", uw(8))))
                           ]))
    assert check(mod)
    serialize_stmt_equal(mod, 'module m :\n'
                              '  input b : UInt<8>\n'
                              '  output a : UInt<8>\n'
                              '\n'
                              '  node n = UInt<1>("h1")\n'
                              '  when n :\n'
                              '    skip\n'
                              '  else :\n'
                              '    a <= b')


def test_module_empty_ports():
    mod = DefModule("m", [], Connect(n("p", uw(8)), u(2, w(8))))
    assert not check(mod)


def test_module_body_wrong():
    mod = DefModule("m", [OutputPort("p", uw(8))],
                    Connect(n("p", uw(8)), s(2, w(8))))
    assert not check(mod)


def test_ext_module_basis():
    mod = DefExtModule("em", [InputPort("b", uw(8)),
                              OutputPort("a", uw(8))], "em")
    assert check(mod)
    serialize_stmt_equal(mod, 'extmodule em :\n'
                              '  input b : UInt<8>\n'
                              '  output a : UInt<8>\n'
                              '\n'
                              '  defname = em')


def test_ext_module_empty_ports():
    mod = DefExtModule("em", [], "em")
    assert not check(mod)
